package javac;

use strict;
use warnings;
use File::Temp qw/tempdir/;
use File::Basename;
use Cwd 'abs_path';

use vars qw/@ISA @EXPORT/;
require Exporter;
@ISA = qw/Exporter/;
@EXPORT = qw/checkSources/;

my $dirname = abs_path ( dirname ( __FILE__ ) ) . "/javac/compile";
sub checkSources {
    # We need source!
    return if @_ != 1;
    my $source = shift;

    # Create a unique temporary directory for us to work in
    my $tmpdir = tempdir ( time . '.XXXX', DIR => "$dirname", CLEANUP => 1 );

    # Create the source file so we can compile it
    open my $FH, ">", "$tmpdir/source.java" or die "Could not open temporary file for compilation: $!\n";
    print $FH $source;
    close $FH;

    # Run javac with Xlint flag, and compile into temporary directory
    open my $JAVAC, "javac -Xlint '$tmpdir/source.java' -d '$tmpdir' 2>&1|";
    my $errors = join "", <$JAVAC>;
    close $JAVAC;
    
    # Errors are "separated" by lines containing a hat character
    my @errorlist = split /\n\s*\^\s*\n/, $errors;

    return [] if @errorlist == 0;

    pop @errorlist; # Because javac shows number of errors on last line

    my $knownTokens = ArjenDB::h -> selectall_hashref ( "SELECT * FROM errors_javac_tokens", "etoken" );
    my $knownRegexp = ArjenDB::h -> selectall_hashref ( "SELECT * FROM errors_javac_regexp", "regexp" );

    my @reports = ();

    foreach my $error ( @errorlist ) {
        my @lines = split "\n", $error;

        # We don't want empty error sets
        next if @lines == 0 or $lines[0] =~ /^\s*$/;

        # First line is:
        # path:lineno: error_token
        $lines[0] =~ /^[^:]+:(\d+):\s*(.*)$/;
        my $line = $1;
        my $etoken = $2;

        next if $etoken =~ /is public, should be declared in a file named/;

        my $report = { line => $line, error_id => undef, more => [], source => $lines [ -1 ], etoken => $etoken };
        shift @lines; pop @lines;
        if ( defined $knownTokens -> { $etoken } ) {
            # The exact error token was found
            runHandler ( $report, $knownTokens -> { $etoken }, \@lines );
        } else {
            # We have to try and match the error token to a known regexp
            foreach my $key ( keys %{ $knownRegexp } ) {
                my $regexp = qr/^\s*$key\s*$/;
                if ( $etoken =~ $regexp ) {
                    # We have found a match!
                    runHandler ( $report, $knownRegexp -> { $key }, \@lines ); 
                    last;
                }
            }
        }

        print STDERR "Unknown error: '" . $report -> { "etoken" } . "'\n" if not defined $report -> { "error_id" };
        print STDERR " occured on line " . $report -> { "line" } . ": '" . $report -> { "source" } . "'\n" if not defined $report -> { "error_id" };
        push @reports, $report;
    }

    return \@reports;
}

sub runHandler {
    return if @_ != 3;
    my $report = shift;
    my $handler = shift;
    my $linesref = shift;
    my @lines = @{ $linesref };

    $report -> { "error_id" } = $handler -> { "error_id" }; 
    if ( defined $handler -> { "handler" } ) {
        # Perl handling code defined
        $report -> { "more" } = eval $handler -> { "handler" };
    }
}

1;
