package miscsyntax;

# This package looks for things like miscased words, e.g. string not String,
# and other simple syntax errors.
#
# (c) 2001, Warren Toomey, GPL3

use strict;
use warnings;
use vars qw/@ISA @EXPORT/;
require Exporter;

@ISA = qw/Exporter/;
@EXPORT = qw/checkSources/;

sub checkSources {
    my $linenum=0;
    my $inComment=0;

    my @errorList;	# List of errors to return

    # Get the database error_id for misspelt word
    my ($misspelt_word_id)= ArjenDB::h -> selectrow_array(
	"select error_id from errors_yacc where etoken = 'misspelt_word';");

    my ($malformed_for_id)= ArjenDB::h -> selectrow_array(
	"select error_id from errors_yacc where etoken = 'malformed_for';");

    my ($bad_comparison_after_andor_id)= ArjenDB::h -> selectrow_array(
	"select error_id from errors_yacc where etoken = 'bad_comparison_after_andor';");

    # List of patterns whose case is often used incorrectly
    my @badcaselist= (
	"System\.out\.println", "System\.out\.print",
	"Scanner", "scan\.nextLine", "scan\.nextDouble",
	"scan\.nextInt", "charAt", "String", "System\.exit", "toString"
    );

    # Get the source code, and
    # work on each line in the source code
    my $source = shift;
    foreach $_ (split(/\n/, $source)) {
      chomp; $linenum++;

      # Get rid of /* ... */ comments which are on 1 line
      s{/\*.*\*/}{}g;

      # If we are in a multi-line comment, try to find the end of it,
      # and keep the stuff after the end of the comment
      if ($inComment && m{\*/(.*)} ) {
        $_= $1; $inComment=0;
      }

      # Still in a multi-line comment, skip it
      next if ($inComment);

      # Look for and remove anything which is the start of a multi-line
      # comment
      if ( m{(.*)/\*} ) {
        $_= $1; $inComment=1;
      }

      # Lose single line comments
      s{//.*}{};

      # Lose strings
      s{"[^"]*"}{}g;

      # Look for misspelt patterns on this line
      foreach my $patt (@badcaselist) {
	if (m{($patt)}i) {
	  my $badspelling= $1;

	  # But if this was actually spelled OK, skip this line
	  next if ($badspelling=~ m{($patt)});

	  # Remove backslashes from the $patt
	  $patt=~ s{\\}{}g;
	  
	  # Create an error and add it to the list
	  my %err= (
		source => $_,
	  	line => $linenum,
		etoken => 'misspelt_word',
		error_id => $misspelt_word_id,
		more => [ $badspelling, $patt ]
	  );

	  push(@errorList, \%err);
        }
      }

      # Look for badly-formed for statements
      if ( m{for\s*\(([^)]*)\)} ) {
	my $insidefor= $1;
	if (! ($insidefor=~ m{^[^;]*;[^;]*;[^;]*$})) {
	  # Create an error and add it to the list
	  my %err= (
		source => $_,
	  	line => $linenum,
		etoken => 'malformed_for',
		error_id => $malformed_for_id,
		more => [ ]
	  );

	  push(@errorList, \%err);
        }
      }

      # Look for && or || followed directly by comparison operators,
      # which usually indicates problems like "if (x > 7 && < 10)
      if ( m{(&&|\|\|)\s*(>|>=|<|<=|==|!=)} ) {
	  my ($first,$second)= ($1,$2);
	  $second=~ s{<}{&lt;}g;
	  $second=~ s{>}{&gt;}g;
	  my %err= (
		source => $_,
	  	line => $linenum,
		etoken => 'bad_comparison_after_andor',
		error_id => $bad_comparison_after_andor_id,
		more => [ $first, $second ]
	  );

	  push(@errorList, \%err);
      }
    }

    return(\@errorList);
}

1;
