package Expressions;
use strict;
use warnings;
use Data::Dumper;

BEGIN {
    use Exporter ();
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    @ISA         = qw(Exporter);
    @EXPORT      = qw(checkOpType expressionType);
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw();
}

use vars @EXPORT_OK;
END { }    # module clean-up code here (global destructor)

### GLOBALS ###

# Here are a list of hashes which hold type groups which can be used
# by certain operators. Note that we use a magic type called 'any'
# which stands in when we cannot deduce the correct type. We ensure that
# all operators can work on 'any' type.

my %enumerabletype = map( { $_ => 1 } qw(byte char short int long any) );
my %numerictype = map( { $_ => 1 }
			qw(byte char short int long float double any) );
my %numstrtype = map( { $_ => 1 }
			qw(byte char short int long float double String any) );
my %booleantype = map( { $_ => 1 } qw(boolean any) );
my %numbooltype = map( { $_ => 1 }
			qw(byte char short int long float double boolean any));


# This hash holds the known operators and the type groups which they
# can work on
my %OpType= (
	'++' => \%enumerabletype,
	'--' => \%enumerabletype,
	'!' => \%booleantype,
	'~' => \%enumerabletype,
	'+' => \%numstrtype,
	'-' => \%numerictype,
	'*' => \%numerictype,
	'/' => \%numerictype,
	'%' => \%enumerabletype,
	'<<' => \%enumerabletype,
	'>>' => \%enumerabletype,
	'>>>' => \%enumerabletype,
	'<' => \%numerictype,
	'>' => \%numerictype,
	'<=' => \%numerictype,
	'>=' => \%numerictype,
	'==' => \%numbooltype,
	'!=' => \%numbooltype,
	'|' => \%enumerabletype,
	'^' => \%enumerabletype,
	'&' => \%enumerabletype,
	'&&' => \%booleantype,
	'||' => \%booleantype,
);

# Some operators return a type which is independent of the operands.
my %OpResult= (
	'<' => 'boolean',
	'>' => 'boolean',
	'<=' => 'boolean',
	'>=' => 'boolean',
	'==' => 'boolean',
	'!=' => 'boolean',
);

# This is a hash of types which can be widened into each other, in
# widening order, i.e. the key is the type name, the value is the
# position in the widening list. Is there a nicer way to code this
# using map()?
my %WidenType;
my $i=0;
foreach my $type ( qw(any byte char short int long float double String) ) {
  $WidenType{$type}= $i;
  $i++;
}


### FUNCTIONS ###


# Given a 3-value arrayref representing an expression, an operator as
# a string, and an second 3-value arrayref representing an expression,
# check that $op can operation on the type of expressions provided.
# Returns 1 if the checks pass, 0 if the checks fail.
# Either $arg1 or $arg2 can be undef, but not both.
sub checkOpType
{
  my ($arg1, $op, $arg2)= @_;

#print("In CheckOpType, op $op\n");
#print("arg1: " . Dumper($arg1));
#print("arg2: " . Dumper($arg2));

  # Panic if unknown $op
  die("Urgh, $op not known in OpType hash\n")
   if (!defined($OpType{$op}));

  return(1);
}

# Given a 3-value arrayref representing an expression, an operator as
# a string, and a 3-value arrayref representing an expression,
# return another 3-value arrayref which represents the expression as a
# string and the type of the expression.
# For example, let's say the input was [ line 23, "x", "int" ], operator +,
# [ line 23, "y", "double" ], the final arrayref would contain
# [ 23, "x + y", "double" ]
#
# We only really deal with numeric types here.
# Either $arg1 or $arg2 can be undef, but not both.
#
sub expressionType
{
  my ($arg1, $op, $arg2)= @_;

  # Build the new expression string which is the combination of the above.
  my $newexpr;
  $newexpr= $arg1->[1] . " " if (defined($arg1));
  $newexpr .= $op;
  $newexpr .= " " . $arg2->[1] if (defined($arg2));

  # Extract both types to make things easier. If one $arg is undef,
  # fill in the type from the other one
  my $linenum= ($arg1) ? $arg1->[0] : $arg2->[0];
  my $a1type= ($arg1) ? $arg1->[2] : $arg2->[2];
  my $a2type= ($arg2) ? $arg2->[2] : $arg1->[2];

  # Check that the operator can work on the given expressions
  if (!checkOpType($arg1, $op, $arg2)) {
    # Not sure what we can do here, as we have definitely found an error.
    # For now, just use the type from one of the arguments as the
    # expression's type.
    return( [ $linenum, $newexpr, $a1type ] );
  }

  # If the operator returns a type independent of the operands, deal with it
  if (defined($OpResult{$op})) {
    return( [ $linenum, $newexpr, $OpResult{$op} ] );
  }

  # Error check: both $a1type and $a2type should be defined here
  if (!defined($a1type)) {
    print(STDERR "a1type not defined!\n");
    print(STDERR Dumper($arg1));
    # For now, set to any
    $a1type = 'any';
  }

  if (!defined($a2type)) {
    print(STDERR "a2type not defined!\n");
    print(STDERR Dumper($arg2));
    # For now, set to any
    $a2type = 'any';
  }

  # Both types the same, use that type.
  if ( $a1type eq $a2type) {
    return( [ $linenum, $newexpr, $a1type ] );
  }

  # Panic if either of the types not listed in %WidenType
  if (!defined($WidenType{$a1type}) || !defined($WidenType{$a2type})) {
    die("Blah, one of $a1type or $a2type not in the WidenType hash\n");
  }

  # Work out the wider type
  my $newtype= ($WidenType{$a2type} > $WidenType{$a1type}) ? $a2type : $a1type;
  return( [ $linenum, $newexpr, $newtype ] );
}

1;
