package Lexer;
use strict;
use warnings;

BEGIN {
    use Exporter ();
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    @ISA         = qw(Exporter);
    @EXPORT      = qw(_Error lexOpen lexClose Lexer);
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw();
}

use vars @EXPORT_OK;
END { }    # module clean-up code here (global destructor)

### GLOBALS ###
our $IN;            # Filehandle for input file
our @TokenQueue;    # Stack of tokens not yet read
our $incomment;     # Are we in a multi-line comment?
our $lnum;          # What line number we are on

# List of keywords in the grammar, and a hash with the keywords as keys.
our @KeywordList = qw{ abstract boolean break byte case catch char class
  continue default do double else extends final
  finally float for if implements import instanceof
  int interface long native new package private protected
  public return short static super switch synchronized
  this throw throws transient try void volatile while
  true false null
};
our %Keyword = map( { $_ => 1 } @KeywordList );

# A few keywords have associated types. This hash lists them
our %KeywordType= (
	'true'  => 'boolean',
	'false' => 'boolean',
);

# The error routine is invoked by the parser when a token is found which
# does not match the rules of the grammar.
sub _Error {

    # Get the list of expected tokens
    my $oktokens= join( ' ', $_[0]->YYExpect );

    # Replace that token with the next one in the stream
    #my ($tok,$val)= Lexer();
    #print Dumper($_[0]);
    #$_[0]->YYCurtok= $tok;
    #$_[0]->YYCurval= $val;
    #$_[0]->YYErrok();
}

# Open the named file and return 1 if OK, 0 otherwise.
sub lexOpen {

    # Open the input file
    open( $IN, "<", $_[0] ) || return (0);
    $incomment = 0;
    $lnum      = 0;
    return (1);
}

sub lexClose {
    close($IN);
}

# The lexer reads from the input file and builds a queue of tokens. When
# called, a single token is read from the queue. Each token consists of
# a string literal (the token proper) and any associated value.
# Returns an empty string for the token when there is no input left.
#
# Each token value is an arrayref with 2 or 3 values. The first two values
# are the line number on which the token occurred and any literal value
# associated with the token. For tokens that have no literal value, the
# token itself is substituted. The optional third value is the type
# associated with the token.
sub Lexer {
    my ($parser) = shift;

    # Loop until we can return a token or there are no lines left
    while (1) {

        # Return a token if possible
        if (@TokenQueue) {
            my $tokptr = shift(@TokenQueue);

            #print("Return token $tokptr->[0] $tokptr->[1]\n");
            return ( @{$tokptr} );
        }

        # No tokens, so read a line in. Exit if no lines left
        $_ = <$IN> || return ( '', undef );
        $lnum++;

        # Remove any // comments and leading whitespace
        s{\s*\/\/.*}{};
        s{^\s*}{};

        # Break the line up into tokens
        while ($_) {
            my $type;    # Holds the deduced type of the token found
            if ($incomment) {

                # If the end of a multi-line comment, note we are out of it
                if (s{.*\*/\s*}{}) { $incomment = 0; next; }

                # Didn't find the comment end, so skip this line
                # as the whole line is part of the comment.
                last;
            }

            # Found the beginning of a multi-line comment, note inside it
            if (s{^/\*\s*}{}) { $incomment = 1; next; }

            # Found either a keyword or an identifier
            if (s{^([A-Za-z_][A-Za-z0-9_]*)\s*}{}) {

                # Push a keyword if we found one
                if ( $Keyword{$1} ) {
                    push(@TokenQueue,[ $1, [ $lnum, $1, $KeywordType{$1} ] ] );
                    next;
                }

                # Otherwise push an identifier. Push its name as the type
                push( @TokenQueue, [ 'IDENTIFIER', [ $lnum, $1, $1 ] ] );
                next;
            }

            # Found a floating-point literal: digits before/after decimal point
            if (s{^((\d+\.\d*|\d*\.\d+)([eE][-+]?\d+)?[fFdD]?)\s*}{}) {
                my $token = $1;
                $type = 'double';
                $type = 'float' if ( $token =~ /[fF]$/ );
                push( @TokenQueue, [ 'FLOATLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Another type of floating-point literal: no decimal point
            if (s{^(\d+[eE][-+]?\d+[fFdD]?)\s*}{}) {
                my $token = $1;
                $type = 'double';
                $type = 'float' if ( $token =~ /[fF]$/ );
                push( @TokenQueue, [ 'FLOATLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Third type of floating-point literal: no decimal point
            if (s{^(\d+([eE][-+]?\d+)?[fFdD])\s*}{}) {
                my $token = $1;
                $type = 'double';
                $type = 'float' if ( $token =~ /[fF]$/ );
                push( @TokenQueue, [ 'FLOATLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Found a hexadecimal integer literal
            if (s{^(0[xX][0-9a-fA-F]*[lL]?)\s*}{}) {
                my $token = $1;
                $type = 'int';
                $type = 'long' if ( $token =~ /[lL]$/ );
                push( @TokenQueue, [ 'HEXINTLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Found an octal integer literal
            if (s{^(0[0-7]*[lL]?)\s*}{}) {
                my $token = $1;
                $type = 'int';
                $type = 'long' if ( $token =~ /[lL]$/ );
                push( @TokenQueue, [ 'OCTINTLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Found a decimal integer literal. Note that we specifically
            # skip leading zeroes, so the single digit 0 is an OCTINTLIT
            if (s{^([1-9]\d*[lL]?)\s*}{}) {
                my $token = $1;
                $type = 'int';
                $type = 'long' if ( $token =~ /[lL]$/ );
                push( @TokenQueue, [ 'DECINTLIT', [ $lnum, $token, $type ] ] );
                next;
            }

            # Found a string literal
            if (s{^"(([^"\\]|\\.)*)"\s*}{}) {
                push( @TokenQueue, [ 'STRINGLIT', [ $lnum, $1, 'String' ] ] );
                next;
            }

            # Found a character literal
            if (s{^'(([^'\\]|\\.))'\s*}{}) {
                push( @TokenQueue, [ 'CHARLIT', [ $lnum, $1, 'char' ] ] );
                next;
            }

            # Found a Unicode character literal
            if (s{^'(\\u[0-9A-Fa-f]{1,4})'\s*}{}) {
                push( @TokenQueue, [ 'CHARLIT', [ $lnum, $1, 'char' ] ] );
                next;
            }

            # Found an octal character literal
            if (s{^'(\\[0-7]{1,3})'\s*}{}) {
                push( @TokenQueue, [ 'CHARLIT', [ $lnum, $1, 'char' ] ] );
                next;
            }

            # Certain multi-character tokens
            if (s{^(>>>=|<<=|>>>|>>=|\+\+|--|&&|\|\||\+=|-=|\*=)\s*}{}) {
                push( @TokenQueue, [ $1, [ $lnum, $1 ] ] );
                next;
            }

            # Certain multi-character tokens
            if (s{^(/=|\%=|\&=|\|=|\^=|==|\!=|<=|>=|<<|>>)\s*}{}) {
                push( @TokenQueue, [ $1, [ $lnum, $1 ] ] );
                next;
            }

            # These 2-char tokens are actually illegal, but
            # we catch them and report them as errors
            if (s{^(=!|=<|=>)\s*}{}) {

                # Replace with the correct token
                my $goodtok = substr( $1, 1, 1 ) . "=";

                push( @TokenQueue, [ $goodtok, [ $lnum, $goodtok ] ] );
                next;
            }

            # Certain single-character tokens
            if (s{^([-+*/&|<>=~.,;:])\s*}{}) {
                push( @TokenQueue, [ $1, [ $lnum, $1 ] ] );
                next;
            }

            # Certain single-character tokens
            if (s{^([\(\)\{\}\[\]\?])\s*}{}) {
                push( @TokenQueue, [ $1, [ $lnum, $1 ] ] );
                next;
            }

            # Whatever is left is unrecognised. Let's keep stuff
            # before any whitespace, trim it off and treat it
            # as the unrecognised token
            s{^(\S+)\s*}{};
        }
    }
}

1;
